/**
 * TerrainMap.java
 * 
 * Representation of the terrain
 * Includes physical topographic information as well as
 * Miscellaneous objects and texturing information
 * 
 * @author mmasukaw
 */

package inf119;
import inf119.Constants;

public class TerrainMap {
	public static final int ROCK = 1, GREENMAN = 2;
		
	
	
	/* Data needed for displaying the map */
	public double resolution;  		//This value represents the distance in meters between each vertex
	public double [][] heights; 	//This is a 2D array of height values at each vertex
									//The map is rendered by connecting the vertices
	public int textureID;			//This corresponds to the TextureManager constants
									//Currently TextureManager.MOON_TEXTURE = 1
									//Unfortunately, we may change how we do this soon...
	
	int [][] objects;				//Miscellaneous objects on the surface of the terrain
									//n x 3 array (n objects which each have 3 pieces of data)
									//ex: objects[1] = {5,7,ROCK}
									//means that the first object is located at (5,7) and is a rock
									//(This convention is subject to change as the team sees fit)
	
	//int [][] landingZones;
							//Format for Internal Landing zone data
							//n x 4 array (n objects which each have 4 pieces of data)
							//ex: landingZones[0] = {4,3, 10, 2000}
	
	LandingZone[] zones;
	
	
	/* Extra data which might be useful to the player. (displayed during pause menu for example?) */
	public String name; //The displayable name of the map, ex: "Titan"
	public String info; //Information about this map, ex: "Saturn VI, the largest moon of Saturn"

	/* 
	 * Extra parameters that may be used in the future (but not yet) 
	 * for random generation of map texture 
	 */
	double[] baseColor; //RGB values for base color (might be used for randomly generated textures
	double[] texture; //The generated texture
	
	public TerrainMap(double res, double[][] hs, int texID, int[][] lzs, int[][] objs)
	//This constructor sets all the data needed for drawing a terrain that uses a fixed texture
	//Example texID = 1 draws a moon texture
	{
		resolution = res;
		if (resolution == 0)
			resolution = 100;
		heights = hs;
		if (heights == null)
			heights = new double[1][1];
		textureID = texID;
		baseColor = new double[3];
		setLandingZones(lzs);
		objects = objs;
		
	}

	private void genTex()
	{
		//fill texture with appropriate data
	}
	
	public void setObjects(int [][] objs)
	{
		objects = objs;
	}
	
	public void setLandingZones(int [][] lzs)
	{
		zones = new LandingZone[lzs.length];
		for (int i = 0; i < lzs.length; i++)
		{
			zones[i] = new LandingZone(lzs[i][0], lzs[i][1], lzs[i][2], lzs[i][3]);
		}
	}
	
	public int[][] getObjects()
	//NOTE: if objects is null -> no objects
	//		returning the pointer allows people to modify the object list directly
	{
		return objects;
	}

	public LandingZone[] getLandingZones()
	{
		return zones;
	}
	
	public boolean setBaseColor(double[] color)
	{
		if (color.length < 3)
			return false;
		else {
			for (int i = 0; i < 3; i++){
				baseColor[i] = color[i];
			}
			genTex();
		}
		return true;
	}
	public double[] getBaseColor()
	{
		return baseColor;
	}
	public double heightAt(double givenX, double givenZ)
	//Returns the height of the vertex closest to {givenX, givenZ}
	{
		int x = (int)Math.round(givenX), z = (int)Math.round(givenZ);
		if (x < 0)
			x = 0;
		if (z < 0)
			z = 0;
		if (x > heights.length - 1)
			x = heights.length - 1;
		if (z > heights[0].length - 1)
			z = heights[0].length - 1;
		
		return heights[x][z];
	}

	private double[] maxHeight(double[] a, double[] b)
	{
		if (b[1] > a[1])
			return b;
		return a;
		            
	}
	private double[] minHeight(double[] a, double[] b)
	{
		if (b[1] < a[1])
			return b;
		return a;
	}
	
	private double interpolateSlope(double x, double z)
	/*
	 * X,Z are cell coords.  Each cell is divided into 2 right triangles like so [/]
	 * Return value is the magnitude of the slope of the line drawn from
	 * the midpoint of the hypotenuse to the other corner of the triangle 
	 */
	{
		int lowX = (int)x, lowZ = (int)z; //The low corner of the cell
		
		if (lowX < 0 || lowZ < 0 || lowX+1 > heights.length-1 || lowZ+1 > heights[0].length-1)
		{
			//If off map: return slope of 0
			//Note: this will cause an out of bounds exception if you have a null height map.
			return 0;
		}
		
		double[] p1 = new double[3], p2 = new double[3], p3 = new double[3], maxH, minH;
		
		//Figure out which is the third point
		if (z < (1-x)) //case: in the triangle with [lowX][lowZ]
		{
			p3[0] = lowX;
			p3[1] = heights[lowX][lowZ];
			p3[2] = lowZ;
		}
		else //case: in the triangle with [lowX+1][lowZ] or on the hypotenuse
		{
			p3[0] = lowX+1;
			p3[1] = heights[lowX+1][lowZ+1];
			p3[2] = lowZ+1;		
		}
		
		p1[0] = lowX+1; p1[1] = heights[lowX][lowZ+1]; p1[2] = lowZ;
		p2[0] = lowX; p2[1] = heights[lowX+1][lowZ]; p2[2] = lowZ+1;
		
		maxH = maxHeight(maxHeight(p1, p2), p3);
		minH = minHeight(minHeight(p1, p2), p3);
		
		if (((maxH[0] == lowX) || (minH[0] == lowX)) && ((maxH[2] == lowZ) || (minH[2] == lowZ))){
			//Case: the high/low pair is the diagonal
			return Math.abs(maxH[1]/Math.sqrt(2));
		}
		else {			
			double length, heightDiff;
			//We know the 2 points are either along the x-axis or the z-axis
			//So we know the difference of the X's == 0 or diff of Z's == 0
			length = (maxH[0] - minH[0]) + (maxH[2] - minH[2]);
			heightDiff = maxH[1] - minH[1];
			return Math.abs(heightDiff/length);
		}
	}
	
	public double interpolateHeight(double givenX, double givenZ)
	/*  
	 * X,Z are cell coords (most likely inbetween cells)
	 * return value is the bilinearly interpolated height of the point at (x,z) in cell units
	 * 
	 * NOTE: this function assumes the Graphics Master draws the terrain map
	 * by dividing each cell like so [/] and drawing the triangles
	 * Thus, the height at any point can be found using bilinear interpolation
	 * But you must first identify which triangle the point is on.  
	 */
	{		
		double x = givenX + (heights.length-1)/2, z = givenZ + (heights[0].length-1)/2; 
		int lowX = (int)x, lowZ = (int)z; //The low corner of the cell
		
		if (lowX < 0 || lowZ < 0 || lowX+1 > heights.length-1  || lowZ+1 > heights[0].length-1)
		{
			//If off map: return height of 0
			//Note: this will cause an out of bounds exception if you have a null height map.
			return 0;
		}
		
		double u, v, w;
		if (z == (1-x)) // case: on the diagonal [/]
		{
			//return the average of the 2 pts on the diagonal
			return (heights[lowX][lowZ+1] + heights[lowX+1][lowZ])/2.0;			
		}			
		else if (z < (1-x)) // case: in the triangle with [lowX][lowZ]
		{
			u = z - lowZ;
			v = x - lowX;
			w = 1 - u - v;
			return (u*heights[lowX+1][lowZ] + v*heights[lowX][lowZ+1] + w*heights[lowX][lowZ]);
		}
		else // case: in the triangle with [lowX+1][lowZ+1]
		{
			u = 1 - (z - lowZ);
			v = 1 - (x - lowX);
			w = 1 - u - v;
			return (u*heights[lowX+1][lowZ] + v*heights[lowX][lowZ+1] + w*heights[lowX+1][lowZ+1]);		
		}
		
		
	}
	
	public boolean safeZone(double x, double z, BoundBox bb, double depthTollerance)
	/*
	 * Inputs X,Y,Z: coordinates in meters, depthTollerance in meters
	 * 
	 * returns true if the lander is not inside the (terrain - tollerance)
	 * returns false if the lander should die
	 */
	{		
		double terrainHeight = resolution * interpolateHeight(x/resolution,z/resolution);
		if (bb.isColliding(terrainHeight-depthTollerance)){
			//The ship is underground, beyond the tollerance range
			return false;
		}
		return true;
	}
	public int isLanding(double x, double z, BoundBox bb, double depthTollerance, double slopeTollerance)
	/*
	 * X,Y,Z are coordinates in meters, depthTollerance in meters, slope some value
	 * 
	 * returns SAFE_LANDING if lander within terrain height +/- tollerance and safe slope
	 * returns NO_LANDING if lander within terrain height +/- tollerance but unsafe slope (or if land on objects)
	 * returns UNSAFE_LANDING if lander not landing 
	 */ 
	{
		double terrainHeight = resolution * interpolateHeight(x/resolution,z/resolution);
		double slope = interpolateSlope(x/resolution,z/resolution);
		if (bb.isColliding(terrainHeight + depthTollerance) == false){
			//Case: too high to land
			return Constants.NO_LANDING;
		}
		else if (bb.isColliding(terrainHeight + depthTollerance) == true){
			//Case: within landing tollerances
			if (slope > slopeTollerance){
				//Case: slope too high, and thus you die
				return Constants.UNSAFE_LANDING;
			}
			return Constants.SAFE_LANDING;
		}
		//Case: you fell through the floor and died
		return Constants.UNSAFE_LANDING;
	}
}